# coding=utf-8
"""
Django settings for repuestazo project.

For more information on this file, see
https://docs.djangoproject.com/en/1.7/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.7/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.7/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'b)b_xfu4+@u5no3b-i&p10g!@n$x590k^jqx-)&4g=kj7c0de7'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False
TEMPLATE_DEBUG = False


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'grimoire.django.xmail',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'xmail_proj.urls'

WSGI_APPLICATION = 'xmail_proj.wsgi.application'

# Database
# https://docs.djangoproject.com/en/1.7/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'repuestazo',
        'USER': 'repuestazo',
        'PASSWORD': 'mysql:repuestazo$2014',
        'HOST': 'localhost',
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.7/topics/i18n/

LANGUAGE_CODE = 'es-ec'
TIME_ZONE = 'America/Guayaquil'
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.7/howto/static-files/

STATIC_URL = 'http://static.repuestazo.com/repuestazo/'
STATIC_ROOT = '/home/rleiva/webapps/static/repuestazo/'
MEDIA_URL = 'http://media.repuestazo.com/repuestazo/'
MEDIA_ROOT = '/home/rleiva/webapps/media/repuestazo/'

# email

try:
    from local_settings import *
except ImportError:
    pass


from ...backends import AsyncEmailBackend
from ...settings import XMAIL_CHUNK_SIZE, XMAIL_BRIDGED_BACKEND
from django.core.management.base import BaseCommand, CommandError
import traceback
from optparse import make_option
from threading import Thread
from time import sleep


class Command(BaseCommand):
    args = 'This command does not expect arguments'
    help = """
    Sends a chunk of e-mails with size as specified in XMAIL_CHUNK_SIZE (defaults to 20) and using the e-mail backend
    specified in XMAIL_BRIDGED_BACKEND (defaults to regular SMTP service).

    You can also pass the --xmail-job-count and --xmail-job-interval if your cron-job tool is not so fine-graned.

    Such options are known to be 1 and 0 by default, which means just one (1) mail-sending loop execution, with zero
    (0) seconds between each execution. As an example, if you want a per-minute execution and your cronjob tool supports
    just a granularity of ten (10) minutes, you should set such options as (10, 60), meaning "10 executions, with a
    distance of 60 seconds between each execution".
    """

    def add_arguments(self, parser):
        parser.add_argument('-c', '--xmail-job-count', action='store', dest='job_count', default=1, type=int,
                            help='number of jobs to run in the same command. By default, 1. Must be >= 1.')
        parser.add_argument('-i', '--xmail-job-interval', action='store', dest='job_count', default=0, type=int,
                            help='delay (in seconds) between jobs. Must be >= 10 if number of jobs is > 0. '
                                 'Defaults to 0. This value is ignored if the number of jobs is set to 1.')

    def handle(self, *args, **options):
        job_count = options.get('job_count', 1)
        job_interval = options.get('job_interval', 0)

        if job_count <= 0:
            raise CommandError("Bad configuration for -c/--xmail-job-count option. A strictly-positive integer "
                               "is expected.")

        if (job_interval < 10) and (job_count > 1):
            raise CommandError("Bad configuration for -i/--xmail-job-interval option. It must be at least of 10 "
                               "seconds if the number if intervals is > 1.")

        def execute():
            try:
                self.stdout.write('Sending a chunk of %d e-mails using backend `%s` ...' % (
                    XMAIL_CHUNK_SIZE, XMAIL_BRIDGED_BACKEND
                ))
                count, total = AsyncEmailBackend.chunk_send()
            except Exception as e:
                raise CommandError('An internal error has occurred when sending the e-mails chunk. See details:\n%s',
                                   traceback.format_exc())
            else:
                self.stdout.write('In-chunk e-mails have been sent: %d out of %d (Check your administration '
                                  'for details)' % (
                    count, total
                ))

        if job_count == 1:
            execute()
        else:
            for _ in xrange(job_count):
                Thread(target=execute).run()
                sleep(job_interval)


from django.db import models
from base64 import encodestring, decodestring
from pickle import loads, dumps
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from .settings import XMAIL_CHUNK_SIZE


class AsyncEmailEntryQuerySet(models.QuerySet):
    """
    Queryset capaz de obtener un subconjunto de sus elementos. Se da preferencia a los que hace mucho no son
      intentados y, en caso de empate, a los que hace mucho hayan sido creados.
    """

    def chunk(self, mark=False):
        queryset = self.exclude(state__in=['succeeded', 'busy']).order_by('tried_on', 'created_on')[0:XMAIL_CHUNK_SIZE]
        if mark:
            fetched_ids = list(queryset.values_list('pk', flat=True))
            queryset = self.filter(pk__in=fetched_ids)
            queryset.update(state='busy')
        return queryset


class AsyncEmailEntry(models.Model):
    """
    Guarda una entrada de mail para ser enviada luego. Los mails con fecha mas "vieja" de ultimo intento seran
      los que primero se intenten enviar.

    Este modelo no nos interesa realmente, salvo que queramos depurar a ver que ha pasado con el envio de los
      correos.
    """

    class CantDecode(Exception):
        pass

    CHOICES = (
        ('pending', _(u'Pending')),
        ('busy', _(u'Busy')),
        ('failed', _(u'Failed')),
        ('succeeded', _(u'Succeeded'))
    )

    # Manager personalizado para bancar otro metodo
    objects = AsyncEmailEntryQuerySet.as_manager()

    # Datos reales
    created_on = models.DateTimeField(default=now, null=False, verbose_name=_(u'Created'),
                                      help_text=_(u"Message creation date (i.e. when send_mail was called on it)"))
    tried_on = models.DateTimeField(default=None, null=True, verbose_name=_(u'Tried'),
                                    help_text=_(u'Last try date'))
    content = models.TextField(null=False, verbose_name=_(u'Pickled e-mail'),
                               help_text=_(u"Pickled message being sent (it is a pickled EmailMessage instance "
                                           u'encoded in base 64)'))
    state = models.CharField(max_length=10, null=False, default='pending', choices=CHOICES,
                             help_text=_(u'Current state of the message (pending, busy, already sent, or failed)'),
                             verbose_name=_(u'Message state'))
    last_error = models.TextField(null=True, editable=False, verbose_name=_(u"Last error"),
                                  help_text=_(u"Details for the last error occurred with this message"))

    # Datos indexados
    to = models.TextField(null=False, verbose_name=_(u"To"),
                          help_text=_(u'Prefetched destinations from the "to" field in the original message'))
    cc = models.TextField(null=True, verbose_name=_(u"Cc"),
                          help_text=_(u'Prefetched destinations from the "cc" field in the original message'))
    bcc = models.TextField(null=True, verbose_name=_(u"Bcc"),
                          help_text=_(u'Prefetched destinations from the "bcc" field in the original message'))
    subject = models.TextField(null=True, verbose_name=_(u"Subject"),
                               help_text=_(u'Prefetched subject from the original message'))

    class Meta:
        verbose_name = _(u'Async e-mail entry')
        verbose_name_plural = _(u'Async e-mail entries')

    def log_exception(self, e):
        """
        Guarda los detalles de la excepcion ocurrida.
        :param e:
        :return:
        """
        self.last_error = "%s : %r" % (type(e).__name__, e.args)

    @property
    def source(self):
        """
        Recupera el mensaje original, o None si hubo un error (la excepcion ocurrida es logueada).
        :return:
        """
        try:
            content = self.content.encode('ascii') if isinstance(self.content, type(u'')) else self.content
            content = content or ''
            return loads(decodestring(content))
        except Exception as e:
            self.log_exception(e)

    @source.setter
    def source(self, value):
        """
        Asigna el mensaje original, si no hubo ningun error (si lo hubo, la excepcion ocurrida es logueada).
        :param value:
        :return:
        """
        try:
            self.content = encodestring(dumps(value))
            self.to = self._comma_break(value.to)
            self.cc = self._comma_break(value.cc)
            self.bcc = self._comma_break(value.bcc)
            self.subject = value.subject
        except Exception as e:
            self.log_exception(e)

    @classmethod
    def build(cls, sources):
        """
        Crea muchas instancias con sus mensajes originales y las guarda.
        :param sources: secuencia de objetos EmailMessage
        :return:
        """
        def instantiate(source):
            instance = cls()
            instance.source = source
            return instance
        if not isinstance(sources, (list, tuple, set, frozenset)):
            sources = [sources]
        return cls.objects.bulk_create([instantiate(source) for source in sources])

    def _comma_break(self, value):
        """
        Une mediante comas los elementos de una secuencia, o devuelve como vino el valor si no es una secuencia.
        :param value:
        :return:
        """
        return u", ".join(value) if isinstance(value, (tuple, list, frozenset, set)) else value

    def _tried(self):
        """
        Actualiza fecha del ultimo intento.
        :return:
        """
        self.tried_on = now()

    def succeeded(self):
        """
        Marca al e-mail como enviado exitosamente.
        :return:
        """
        self._tried()
        self.state = 'succeeded'

    def failed(self):
        """
        Marca al e-mail como fallado al enviarse.
        :return:
        """
        self._tried()
        self.state = 'failed'


from django.contrib.admin import site, ModelAdmin
from django.utils.translation import ugettext_lazy as _
from .models import AsyncEmailEntry


class AsyncMailAdmin(ModelAdmin):
    """
    No permite editar ni crear pero permite borrar. Muestra un subconjunto de los campos.
    """

    list_display = ['created_on', 'tried_on', 'get_state', 'to', 'cc', 'bcc', 'subject', 'last_error']
    list_display_links = None

    def get_state(self, obj):
        return obj.get_state_display()
    get_state.short_description = _(u'Message state')

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        # permitimos que, si no tenemos un objeto seleccionado,
        # podamos listarlo
        return obj is None

    def has_delete_permission(self, request, obj=None):
        return True


site.register(AsyncEmailEntry, AsyncMailAdmin)


from django.conf import settings


# How many e-mails can be sent at a given time
XMAIL_CHUNK_SIZE = getattr(settings, 'XMAIL_CHUNK_SIZE', 20)

# The actual e-mail backend which will send the e-mails
XMAIL_BRIDGED_BACKEND = getattr(settings, 'XMAIL_BRIDGED_BACKEND', 'django.core.mail.backends.smtp.EmailBackend')


from django.core.exceptions import ImproperlyConfigured
from django.db.transaction import atomic
from django.utils.module_loading import import_string
from .settings import XMAIL_BRIDGED_BACKEND
from .models import AsyncEmailEntry


class AsyncEmailBackend(object):
    """
    Este backend no envia los mails sino que, en realidad, los guarda en una base de datos.

    Para utilizar este backend, deberiamos usar, en nuestros settings, algo como:

    EMAIL_BACKEND = 'xmail.backends.AsyncEmailBackend'
    XMAIL_BRIDGED_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
    XMAIL_CHUNK_SIZE = 20

    En el caso de XMAIL_BRIDGED_BACKEND, el valor predeterminado es django.core.mail.backends.smtp.EmailBackend,
      por lo que es opcional asignar esa configuracion si es la que se va a utilizar.

    En el caso de XMAIL_CHUNK_SIZE, el valor predeterminado es 20, por lo que es opcional asignarle valor si nos
      sentimos comodos con esa cantidad. Este valor indica la cantidad de mails que se van a procesar por iteracion.

    En algun momento, vamos a tener que enviar los correos guardados en la base de datos, y para esto
      se utiliza una utilidad en este mismo backend (pero sin instanciarlo). Dicha utilidad va a hacer
      uso de un backend regular de mails (uno sincrono), pero esto ocurre en el contexto de una operacion
      externa (es decir: dicha utilidad sera invocada desde un comando de manage.py).

    En cuanto a las configuraciones relativas a cada backend, hay que especificarlas normalmente. Esta configuracion
      es la que se va a utilizar para realizar el envio del mail dentro de la utilidad, de la misma forma en como
      ocurriria normalmente, pero refiriendose a XMAIL_BRIDGED_BACKEND en lugar de a EMAIL_BACKEND.

    Cabe aclarar que NO tendremos control sobre la conexion utilizada cuando los correos sean finalmente enviado, como
      SI lo tenemos cuando enviamos el mail en una invocacion a send_mail (mas que nada por el hecho de que, para esto,
      cada mail deberia tener una conexion diferente segun se le especifique).
    """

    def __init__(self, **kwargs):
        """
        En realidad no deberiamos hacer nada con estas opciones. No parece que las vayamos a usar.
        :param kwargs:
        :return:
        """
        self.options = kwargs

    def open(self):
        """
        No hace nada ya que no se conecta a ningun lado.
        :return:
        """
        pass

    def close(self):
        """
        No hace nada ya que no se conecta a ningun lado.
        :return:
        """
        pass

    def send_messages(self, email_messages):
        """
        Construye y guarda los e-mails. No los envia en este momento.
        :param email_messages:
        :return:
        """
        return len(AsyncEmailEntry.build(email_messages))

    @classmethod
    def chunk_send(cls):
        """
        Realiza el envio real de los mails. Este debe, en realidad, invocarse desde un comando de manage.py.

        El envio de e-mail se hace mediante la instanciacion de algun otro back-end que DEBEMOS
        :return:
        """

        # Debemos tener XMAIL_BRIDGED_BACKEND en el server. Este debe apuntar a un path correcto que seria un
        # backend de e-mail que usariamos regularmente si no hicieramos envios asincronos.
        #
        # Por esto de que el backend debe ser uno sincrono (regular), no se le puede especificar a esta propiedad
        # un backend asincrono que dependa de XMAIL_BRIDGED_BACKEND ya que causaria una recursividad infinita.
        try:
            klass = import_string(XMAIL_BRIDGED_BACKEND)
        except:
            raise ImproperlyConfigured("For this backend to work, XMAIL_BRIDGED_BACKEND must be set to a valid "
                                       "python path for a normal, synchronous, e-mail backend (by default, it is the "
                                       "SMTP backend provided by Django)")

        if issubclass(klass, cls):
            raise ImproperlyConfigured("XMAIL_BRIDGED_BACKEND cannot be an asynchronous mail backend "
                                       "(you used `%s`. Perhaps you wanted to set that value to "
                                       "EMAIL_BACKEND?)" % XMAIL_BRIDGED_BACKEND)

        # Creamos un backend, y mandamos un lote de correos (preferentemente nuevos). Este lote de correos se puede
        # configurar tranquilamente con un setting. De manera predeterminada, en cada vuelta se intentan enviar 20
        # correos.
        backend = klass(fail_silently=False)

        try:
            count = 0
            backend.open()
            chunk = AsyncEmailEntry.objects.chunk(mark=True)
            total = len(chunk)
            for entry in chunk:
                with atomic():
                    # Que esto se nos ejecute dentro de una transaccion es buena idea ya que
                    # independientemente de las otras cosas que ocurran, este mail ya se habra
                    # enviado al destinatario.
                    #
                    # Si el mensaje no se puede despicklear, entonces se guarda un error asociado
                    # a esa operacion. En caso contrario, se intenta enviar el mail.
                    source = entry.source
                    if source:
                        try:
                            if backend.send_messages([source]):
                                entry.succeeded()
                                count += 1
                            else:
                                entry.last_error = "send_messages() could not send it"
                                entry.failed()
                        except Exception as e:
                            entry.log_exception(e)
                            entry.failed()
                    else:
                        entry.failed()
                    entry.save()
            backend.close()
            return count, total
        except:
            raise


from setuptools import setup, find_packages

setup(
    name='django-xmail-ritual',
    version='0.0.13',
    namespace_packages=['grimoire', 'grimoire.django'],
    packages=find_packages(exclude=['xmail_proj', 'xmail_proj.*']),
    package_data={
        'grimoire.django.xmail': [
            'locale/*/LC_MESSAGES/*.*'
        ]
    },
    url='https://github.com/luismasuelli/django-xmail-ritual',
    license='LGPL',
    author='Luis y Anita',
    author_email='luismasuelli@hotmail.com',
    description='The xmail library is a mail dispatching library for Django 1.7. Mails are sent asynchronously by a '
                'management command',
    install_requires=['python-cantrips>=0.6.6', 'Django>=1.7']
)


from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class DefaultAppConfig(AppConfig):
    name = 'grimoire.django.xmail'
    verbose_name = _(u'Asynchronous mailing')


# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.db import models, migrations
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AsyncEmailEntry',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_on', models.DateTimeField(default=django.utils.timezone.now, help_text='Message creation date (i.e. when send_mail was called on it)', verbose_name='Created')),
                ('tried_on', models.DateTimeField(default=None, help_text='Last try date', null=True, verbose_name='Tried')),
                ('content', models.TextField(help_text='Pickled message being sent (it is a pickled EmailMessage instance encoded in base 64)', verbose_name='Pickled e-mail')),
                ('state', models.CharField(default=b'pending', help_text='Current state of the message (pending, busy, already sent, or failed)', max_length=10, verbose_name='Message state', choices=[(b'pending', 'Pending'), (b'busy', 'Busy'), (b'failed', 'Failed'), (b'succeeded', 'Succeeded')])),
                ('last_error', models.TextField(help_text='Details for the last error occurred with this message', verbose_name='Last error', null=True, editable=False)),
                ('to', models.TextField(help_text='Prefetched destinations from the "to" field in the original message', verbose_name='To')),
                ('cc', models.TextField(help_text='Prefetched destinations from the "cc" field in the original message', null=True, verbose_name='Cc')),
                ('bcc', models.TextField(help_text='Prefetched destinations from the "bcc" field in the original message', null=True, verbose_name='Bcc')),
                ('subject', models.TextField(help_text='Prefetched subject from the original message', null=True, verbose_name='Subject')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))


ALLOWED_HOSTS = []


DEBUG = True
TEMPLATE_DEBUG = True


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')
MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')